using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Net;
using System.Net.NetworkInformation;
using System.Net.Sockets;
using System.IO;

using System.Text.Json;
using System.Text.Json.Serialization;

namespace TauPlugin
{
    public class TauNetwork
    {

        public int connectionStatus { get; private set; } = 0; 
        bool reconnectPending = false;

        public UdpClient thetaClient = new UdpClient();
        public IPEndPoint remoteEP = new IPEndPoint(0, 0);
        private Thread mainThread;
        const ushort announcePort = 16061;
        const string thetaNetName = "ttracker.local";
        //const string thetaNetName = "10.0.0.5";

        const ushort thetaPort = 19920;

        //byte[] rcvBuf;
        BinaryReader rcvBufReader = new BinaryReader(new MemoryStream());
        BinaryReader packetBufReader = new BinaryReader(new MemoryStream());

        private TauObjects.DataPacket datapacket = new TauObjects.DataPacket();
        private TauObjects.DataPacket datapacket_temp = new TauObjects.DataPacket();
        private TauObjects.DebugPacket debugpacket = new TauObjects.DebugPacket();
        private TauObjects.DebugPacket debugpacket_temp = new TauObjects.DebugPacket();
        private TauObjects.MappingPacket mappingpacket = new TauObjects.MappingPacket();

        private TauObjects.AnnouncerDataObj announce_packet;

        public object DataLocker = new object(),
                    MappingLocker = new object(),
                    LogsLocker = new object(),
                    DebugLocker = new object();

        bool pageLost = false;
        int prevPacketId = 0;
        int packetId = 0;
        int pagesTotal = 0;
        int prevPage = 0;
        int pageCur = 0;

        public void parsePacket(byte[] packet)
        {
            rcvBufReader.BaseStream.SetLength(0);                       // reset stream
            rcvBufReader.BaseStream.Write(packet, 0, packet.Length);    // write packet
            rcvBufReader.BaseStream.Position = 0;

            byte b1 = rcvBufReader.ReadByte();
            byte b2 = rcvBufReader.ReadByte();

            // confirmation packet
            if (b1 == 'o' && b2 == 'k')
            {
                return;
            }

            // data packet
            if (b1 != 'd' || b2 != 'p')
            {
                System.Console.WriteLine("bad packet");
                return;
            }

            //read header
            packetId = rcvBufReader.ReadUInt16();
            pageCur = rcvBufReader.ReadByte();
            pagesTotal = rcvBufReader.ReadByte();

            //StringBuilder strBld = new StringBuilder(32);
            //strBld.AppendFormat("p_id: {0:x2}; t: {1}; c: {2}", packetId, pagesTotal, pageCur);
            //System.Console.WriteLine(strBld);

            if (packetId != prevPacketId)
            {
                //System.Console.WriteLine("new packet!");
                pageLost = false;
                prevPage = -1;
                packetBufReader.BaseStream.SetLength(0);    // clear packet buffer
            }
            prevPacketId = packetId;
            packetBufReader.BaseStream.Write(packet, 6, packet.Length - 6); // write page to packet buffer, skip page header

            if (pageCur != prevPage + 1)
            {
                pageLost = true;
                System.Console.WriteLine("page lost!");
            }
            prevPage = pageCur;

            if (pagesTotal == pageCur + 1 && !pageLost)
            {
                packetBufReader.BaseStream.Position = 0;    // rewind

                byte[] opts = { 0, 0, 0 };
                ushort[] szs = { 0, 0, 0 };

                //System.Console.WriteLine("parsing packet");
                for (int i = 0; i < 3; i++)
                {
                    opts[i] = packetBufReader.ReadByte();
                    //System.Console.Write("" + opts[i] + " ");
                }
                //System.Console.WriteLine();

                for (int i = 0; i < 3; i++)
                {
                    if (opts[i] > 0)
                    {
                        szs[i] = packetBufReader.ReadUInt16();
                    }
                    else
                    {
                        szs[i] = 0;
                    }
                    //System.Console.Write("" + szs[i] + " ");
                }
                //System.Console.WriteLine();

                // parse data block
                if (opts[0] > 0)
                {
                    //parseData(packet_buf);
                    if (mappingpacket != null && mappingpacket.initialized)
                    {
                        lock (MappingLocker)
                        {
                            datapacket_temp.ParseUpdate(packetBufReader, mappingpacket);
                        }
                    }
                    else {
                        datapacket_temp.ParseUpdate(packetBufReader);
                    }


                    lock (DataLocker) {
                        datapacket.CopyFrom(datapacket_temp);
                        //System.Console.WriteLine(datapacket.ToString());
                    }
                }
                // parse data_ext block
                if (opts[1] > 0)
                {
                    //parseDataExt(packet_buf);
                }
                // parse mapping block
                if (opts[2] > 0)
                {
                    //parseMapping(packet_buf, szs[2]);
                    lock (MappingLocker) {
                        mappingpacket.Parse(packetBufReader.ReadBytes(szs[2]));
                    }
                }
            }
        }


        public void OnDataCallback(IAsyncResult result)
        {
            UdpClient socket = result.AsyncState as UdpClient;
            
            try
            {
                byte[] data = socket.EndReceive(result, ref remoteEP);

                parsePacket(data);

                //Console.WriteLine("Got " + message.Length + " bytes from " + source);
            }
            catch (System.Net.Sockets.SocketException) {
                //Console.WriteLine("TauNetwork: disconnected, can't get data");
                reconnectPending = true;
            }
            socket.BeginReceive(new AsyncCallback(OnDataCallback), socket);
        }


        public void OnAnnounceCallback(IAsyncResult result)
        {
            UdpClient socket = result.AsyncState as UdpClient;
            IPEndPoint source = new IPEndPoint(0, 0);
            byte[] announce = socket.EndReceive(result, ref source);
            string announce_str = Encoding.UTF8.GetString(announce, 0, announce.Length);
            announce_packet = JsonSerializer.Deserialize<TauObjects.AnnouncerDataObj>(announce_str);
            announce_packet.HubIP = source.Address.ToString();
            //Console.WriteLine(announce_packet.HubIP);

            socket.BeginReceive(new AsyncCallback(OnAnnounceCallback), socket);
        }

        public TauObjects.DataPacket GetDataPacket()
        {
            if (datapacket != null && datapacket.initialized)
            {
                return datapacket;
            }
            else {
                return null;
            }
            
        }

        private void RespawnClient(){
            thetaClient = new UdpClient();
            thetaClient.Client.SendTimeout = 2000;
            thetaClient.Client.ReceiveTimeout = 2000;
        }

        private void ConnectInitialization() {
            IPEndPoint announceAddress = new IPEndPoint(IPAddress.Any, announcePort);
            UdpClient announceClient = new UdpClient();
            announceClient.Client.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true);
            announceClient.ExclusiveAddressUse = false;
            announceClient.Client.Bind(announceAddress);
            announceClient.BeginReceive(new AsyncCallback(OnAnnounceCallback), announceClient);

            IPEndPoint ep_ = new IPEndPoint(0, 0);
            // drq - data request
            // 1 - data 
            // 1 - ext data (debug)
            // 1 - mapping, mapping will be sent once
            
            byte[] rqPacket = Encoding.ASCII.GetBytes("drq101");

            RespawnClient();
            while (!thetaClient.Client.Connected){

                try { //try domain connection
                    RespawnClient();
                    //Console.WriteLine("trying domain connection");
                    thetaClient.Connect(thetaNetName, thetaPort);
                    thetaClient.Send(rqPacket, rqPacket.Length);
                    var TODO_CHECK_THAT_ANSWER_IS_OK = thetaClient.Receive(ref ep_);
                } 
                catch (System.Net.Sockets.SocketException){ //fallback to announcer data
                    if (announce_packet != null) {
                        try {
                            RespawnClient();
                            //Console.WriteLine("trying announcer connection");
                            thetaClient.Connect(announce_packet.HubIP, thetaPort);
                            thetaClient.Send(rqPacket, rqPacket.Length);
                            var TODO_CHECK_THAT_ANSWER_IS_OK = thetaClient.Receive(ref ep_);
                        } 
                        catch (System.Net.Sockets.SocketException){ //well, fuck.
                            //Console.WriteLine("connection didn't worked at all");
                        }
                    }
                    else
                    {
                        //Console.WriteLine("announce_packet null");
                    }

                }
                //Console.WriteLine(thetaClient.Client.Connected);
                Thread.Sleep(100);
            }
            

            thetaClient.BeginReceive(new AsyncCallback(OnDataCallback), thetaClient);
        }
        public void MainThreadTask(){

            byte[] rqPacket = Encoding.ASCII.GetBytes("drq101");

            ConnectInitialization();

            while (connectionStatus == 1)
            {
                if (reconnectPending) {
                    reconnectPending = false;
                    ConnectInitialization();
                }
                else {
                    thetaClient.Send(rqPacket, rqPacket.Length);
                }
                
                Thread.Sleep(10000);
                //System.Console.WriteLine(thetaClient.Client.Connected);
            }
        }

        public int Connect()
        {
            if (connectionStatus == 0)
            {
                connectionStatus = 1;
                mainThread = new Thread(() => MainThreadTask()) { IsBackground = true };
                mainThread.Start();
                return 0; //success
            }
            else if (connectionStatus == 1)
            {
                return 1; //can't connect, already connected state
            }
            else if (connectionStatus == 2)
            {
                return 2; //can't connect, disconnect pending
            }
            else
            {
                throw new Exception("TauNetwork.Connect: ConnectionStatus unexpected value");
            }
        }

        public int Disconnect()
        {
            if (connectionStatus == 1) {
                connectionStatus = 2;
                while (mainThread.IsAlive) {
                    Thread.Sleep(5);
                }

                connectionStatus = 0;
                return 0;
            }
            else if (connectionStatus == 0)
            {
                return 1; //can't disconnect, not connected state
            }
            else if (connectionStatus == 2)
            {
                return 2; //can't disconnect, disconnect pending
            }
            else
            {
                throw new Exception("TauNetwork.Disconnect: ConnectionStatus unexpected value");
            }
        }
    }
}